#rm(list=ls(all=TRUE))
library(VGAM)
library(mclust)
library(EMMIXskew)
library(clusteval) 

fGH.BS=function(x,d,mu,tau,sigma,kappa,chi,psi){
  xprod=c(chi+diag(t(x-mu)%*%solve(sigma)%*%(x-mu)))
  xtau=c(t(x-mu)%*%solve(sigma)%*%(tau))
  tauprod=c(psi+tau%*%solve(sigma)%*%(tau))
  part1=besselK(sqrt(tauprod*xprod),kappa-(d/2))/besselK(sqrt(chi*psi),kappa)
  part2=(tauprod/sqrt(tauprod*xprod))^((d/2)-kappa)
  ex=(psi/chi)^(kappa/2)*exp(xtau)/((det(sigma))^0.5*(2*pi)^(d/2))
  return(part1*part2*ex)
}
f.BS=function(x,d,mu,tau,sigma,alpha){
  0.5*(fGH.BS(x,d,mu,tau,sigma,.5,alpha^(-2),alpha^(-2))+
         fGH.BS(x,d,mu,tau,sigma,-.5,alpha^(-2),alpha^(-2)))    
}

f=function(x,d,mu,tau,sigma,alpha,p){
  ff=0
  for(i in 1:length(p)){
    ff=ff+p[i]*f.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],alpha[i])
  }
  return(ff)
}

#### EM function ####
MIXNMV.BS.EM=function (x, mu, tau, sigma, alpha, p, tol=1e-4, max.iter=100, per, class=c(), Erro=F)
{
  begin = proc.time()[1]
  n=dim(x)[2]
  d=dim(x)[1]
  g=length(p)
  
  m=length(p)*d*(d+1)/2+length(alpha)+length(p)-1+2*length(p)*length(mu[[1]])
  ind.den=f(x,d,mu,tau,sigma,alpha,p)
  lk=logli.old=sum(log(ind.den))
  iter=0
  cat("\t\t FM-NMVBS-EM (g = ",g,")\n",sep="")
  cat(paste(rep("=",60), sep="", collapse=""),"\n")
  cat("iter =", iter, "\t logli.old=",  logli.old, "\n")
  
  R=function(cc,k,a) besselK(cc,k+a)/besselK(cc, k)
  mu.n=sigma.n=tau.n=list()
  repeat
  {
    iter = iter+1
    ### E step
    z=w=t=matrix(0,g,n)
    
    for(i in 1:g){
      p.x=fGH.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],.5,alpha[i]^(-2),alpha[i]^(-2))/
        (fGH.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],.5,alpha[i]^(-2),alpha[i]^(-2))+
           fGH.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],-.5,alpha[i]^(-2),alpha[i]^(-2)))
      psii=c(1/alpha[i]^2+tau[[i]]%*%solve(sigma[[i]])%*%(tau[[i]]))
      chiij=c(1/alpha[i]^2+diag(t(x-mu[[i]])%*%solve(sigma[[i]])%*%(x-mu[[i]])))
      cc=sqrt(chiij*psii)
      w[i,]=sqrt(chiij/psii)*(p.x*R(cc,(1-d)/2,1)+(1-p.x)*R(cc,-(d+1)/2,1)) 
      t[i,]=sqrt(psii/chiij)*(p.x*R(cc,(d-1)/2,1)+(1-p.x)*R(cc,(d+1)/2,1))
      z[i,] = p[i]*f.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],alpha[i])
    }
    z=z/matrix(apply(z,2,sum),g,n,byrow = T)
    ni = apply(z,1,sum)
    Ai = apply(z*t,1,sum)
    Ci = apply(w*z,1,sum)
    S = w+t-2
    
    ### step 1
    p.n = ni/n
    alpha.n = sqrt(apply(S*z,1,sum)/ni)
    
    Xbar=z%*%t(x)
    ### step 3
    #Bi = list()
    for(i in 1:g){
      Bi = apply(x*matrix(z[i,],d,n,byrow=T)*matrix(t[i,],d,n,byrow=T),1,sum)
      ###lambda
      tau.n[[i]] = (Ai[i]*apply(x*matrix(z[i,],d,n,byrow=T),1,sum)-
                    ni[i]*Bi)/(Ai[i]*Ci[i]-ni[i]^2)
      mu.n[[i]] = (Bi-ni[i]*tau.n[[i]])/Ai[i]
      
      t.ro = matrix(c(numeric(d)),d,d)
      for(k in 1:n){
        t.ro = t.ro+t[i,k]*z[i,k]*((x[,k]-mu.n[[i]])%*%t(x[,k]-mu.n[[i]]))
      }
      sigma.n[[i]] = ((t.ro)/ni[i]+(tau.n[[i]]%*%t(tau.n[[i]]))*Ci[i]/ni[i]-
                      tau.n[[i]]%*%t(Xbar[i,]/ni[i]-mu.n[[i]])-
                      (Xbar[i,]/ni[i]-mu.n[[i]])%*%t(tau.n[[i]]))  
    }
    
    ind.den=f(x,d,mu.n,tau.n,sigma.n,alpha.n,p.n)
    logli.new=sum(log(ind.den))
    # Aikten's method
    lk = c(lk, logli.new)
    if (iter < 2) epsilon = abs(logli.old-logli.new)/abs(logli.new)
    else 
    {
      tmp  = (logli.new - logli.old)/(logli.old-lk[iter-1])
      tmp2 =  logli.old + (logli.new-logli.old)/(1-tmp)
      epsilon = abs(tmp2-logli.new)
    }
    diff = logli.new -logli.old
    if(iter %% per == 0 | is.na(diff))
    {
      cat('iter =', iter, '\t logli =', logli.new, '\t diff =', diff, "\t Aikten's diff =", epsilon, '\n')
    }
    if(is.na(diff)) {logli.new=logli.old 
    break}
    if(is.na(epsilon)) {logli.new=logli.old 
    break}
    if(epsilon < tol | iter == max.iter) break
    if(diff < tol | iter == max.iter) break
    {logli.old = logli.new
    alpha=alpha.n
    mu=mu.n
    sigma=sigma.n
    tau=tau.n
    p=p.n
    }
  }
  BIC = -2*logli.new+m*log(n)
  AIC = 2*(-logli.new+m)
  z = matrix(0,g,n)
  for(i in 1:g){
    z[i,] = p[i]*f.BS(x,d,mu[[i]],tau[[i]],sigma[[i]],alpha[i])
  }
  ZZ=t(z)
  z=z/matrix(apply(z,2,sum),g,n,byrow = T)
  ent = -sum(z*log(z+1e-10))
  ICL = BIC+2*ent
  CLC=2*(-logli.new+ent)
  EDC=-2*logli.new+0.2*m*sqrt(n)
  
  #ZZ=matrix(numeric(n*length(p)),n,length(p))
  #for(f in 1:length(p)){
  # ZZ[,f]=c(p[f] *f.BS(x,d,mu[[f]],tau[[f]],sigma[[f]],alpha[f]))
  #}
  ind.den = rowSums(ZZ)
  ZZ = ZZ / ind.den; 
  post.cate = matrix(apply(ZZ, 1, order), length(p))[length(p),]
  di = table(post.cate)
  
  cat(paste(rep("=",60),sep="",collapse=""),"\n")
  cat('iter =', iter, '\tlogli =', logli.new, '\t diff =', diff, '\n')

  end = proc.time()[1]
  cat('NMVBSmix takes', end - begin, 'seconds\n\n')
  
  para = list(p=p, mu = mu , tau=tau, sigma = sigma , alpha = alpha )
  if(is.numeric(class)==T){
    true.clus=class
    CCR=1-error.rate(post.cate, true.clus)
    #RI=rand.index(post.cate, true.clus)
    ARI=adjustedRandIndex(post.cate, true.clus)
    jacc_ind=cluster_similarity(true.clus, post.cate)
    
    out=list(logli = logli.new, para = para, ni = di, post.cate = post.cate, 
             para.len = m,AIC = AIC, BIC = BIC, ICL = ICL,EDC = EDC , CLC = CLC,
             CCR = CCR, ARI=ARI, jacc_ind = jacc_ind)
  }
  else{out = list(iter=iter ,logli = logli.new, para = para, ni = di, post.cate = post.cate, 
                  para.len = m,AIC = AIC, BIC = BIC, ICL = ICL,EDC = EDC , CLC = CLC)}
  
  #mds=c(logli=logli.new,num.para=m, AIC=AIC, BIC=BIC, ENT=ent, ICL=ICL, ARI=ARI, CCR=CCR)
  mds=c(logli=logli.new, AIC=AIC, BIC=BIC,  ARI=ARI, CCR=CCR)
  cat("model information:\n"); print(mds); cat("\n")
  return(list(mds=mds, para=para))
}