#rm(list=ls(all=TRUE))
#library(VGAM)
library(mclust)
library(EMMIXskew)
library(clusteval)

fSAL=function(x,d,mu,lambda,sigma){
  nu=(2-d)/2
  sig.siv=solve(sigma)
  a=2+c((lambda)%*%sig.siv%*%lambda)
  u=sqrt(a*diag(t(x-mu)%*%sig.siv%*%(x-mu)))
  ex=c(2*exp(t(x-mu)%*%sig.siv%*%lambda)/((det(sigma))^0.5*(2*pi)^(d/2)))
  part1= besselK(u,nu)*diag(t(x-mu)%*%sig.siv%*%(x-mu)/a)^(nu/2)
  return(part1*ex)
}
f.mix=function(x,d,mu,lambda,sigma,p){
  ff=0
  for(i in 1:length(p)){
    ff=ff+p[i]*fSAL(x,d,mu[[i]],lambda[[i]],sigma[[i]])
  }
  return(ff)
}
#### EM function ####
MIXSAL.EM=function (x, mu, lambda, sigma, p, tol=1e-5, max.iter=100, per, class=c())
{
  begin = proc.time()[1]
  library(HyperbolicDist)
  n=dim(x)[2]
  d=dim(x)[1]
  g=length(p)
  
  m=length(p)*d*(d+1)/2+length(p)-1+2*length(p)*length(mu[[1]])
  ind.den=f.mix(x,d,mu,lambda,sigma,p)
  lk=logli.old=sum(log(ind.den))
  iter=0
  #cat("iter =", iter, "\t logli.old=",  logli.old, "\n")
  
  RR=function(cc,k,a)  besselRatio(cc,k,a, useExpScaled = 500) 
  R=Vectorize(RR)
  mu.n=sigma.n=lambda.n=list()
  repeat
  {
    iter = iter+1
    ### E step
    z=w=t=matrix(0,g,n)
    
    for(i in 1:g){
      nu=(2-d)/2
      psii=2+lambda[[i]]%*%solve(sigma[[i]])%*%(lambda[[i]])
      chiij=diag(t(x-mu[[i]])%*%solve(sigma[[i]])%*%(x-mu[[i]]))
      cc=sqrt(chiij*psii)
      w[i,]=sqrt(chiij/psii)*R(cc,nu,1) 
      t[i,]=sqrt(psii/chiij)*R(cc,nu,1)-2*nu/chiij
      z[i,] = p[i]*fSAL(x,d,mu[[i]],lambda[[i]],sigma[[i]])
    }
    z=z/matrix(apply(z,2,sum),g,n,byrow = T)
    ni = apply(z,1,sum)
    Ai = apply(z*t,1,sum)
    Ci = apply(w*z,1,sum)
    
    ### step 1
    p.n = ni/n
    
    Xbar=z%*%t(x)
    ### step 3
    for(i in 1:g){
      Bi = apply(x*matrix(z[i,],d,n,byrow=T)*matrix(t[i,],d,n,byrow=T),1,sum)
      ###lambda
      lambda.n[[i]] = (Ai[i]*apply(x*matrix(z[i,],d,n,byrow=T),1,sum)-
                         ni[i]*Bi)/(Ai[i]*Ci[i]-ni[i]^2)
      mu.n[[i]] = (Bi-ni[i]*lambda.n[[i]])/Ai[i]
      
      t.ro = matrix(c(numeric(d)),d,d)
      for(k in 1:n){
        t.ro = t.ro+t[i,k]*z[i,k]*((x[,k]-mu.n[[i]])%*%t(x[,k]-mu.n[[i]]))
      }
      sigma.n[[i]] = ((t.ro)/ni[i]+(lambda.n[[i]]%*%t(lambda.n[[i]]))*Ci[i]/ni[i]-
                        lambda.n[[i]]%*%t(Xbar[i,]/ni[i]-mu.n[[i]])-
                        (Xbar[i,]/ni[i]-mu.n[[i]])%*%t(lambda.n[[i]]))  
    }
    
    ind.den=f.mix(x,d,mu.n,lambda.n,sigma.n,p.n)
    logli.new=sum(log(ind.den))
    # Aikten's method
    lk = c(lk, logli.new)
    if (iter < 2) epsilon = abs(logli.old-logli.new)/abs(logli.new)
    else 
    {
      tmp  = (logli.new - logli.old)/(logli.old-lk[iter-1])
      tmp2 =  logli.old + (logli.new-logli.old)/(1-tmp)
      epsilon = abs(tmp2-logli.new)
    }
    diff = logli.new -logli.old
    
    if(is.na(diff)) {logli.new=logli.old 
    break}
    if(is.na(epsilon)) {logli.new=logli.old 
    break}
    if(epsilon < tol | iter == max.iter) break
    if(diff < tol | iter == max.iter) break
    {logli.old = logli.new
    mu=mu.n
    sigma=sigma.n
    lambda=lambda.n
    p=p.n
    }
  }
  BIC = -2*logli.new+m*log(n)
  AIC = 2*(-logli.new+m)
  z = matrix(0,g,n)
  for(i in 1:g){
    z[i,] = p[i]*fSAL(x,d,mu[[i]],lambda[[i]],sigma[[i]])
  }
  ZZ=t(z)
  z=z/matrix(apply(z,2,sum),g,n,byrow = T)
  ent = -sum(z*log(z))
  ICL = BIC+2*ent
  CLC=2*(-logli.new+ent)
  EDC=-2*logli.new+0.2*m*sqrt(n)
  
  #ZZ=matrix(numeric(n*length(p)),n,length(p))
  #for(f in 1:length(p)){
  # ZZ[,f]=c(p[f] *f.BS(x,d,mu[[f]],lambda[[f]],sigma[[f]],alpha[f]))
  #}
  ind.den = rowSums(ZZ)
  ZZ = ZZ / ind.den; post.cate = matrix(apply(ZZ, 1, order), length(p))[length(p),]
  di = table(post.cate)
  
  end = proc.time()[1]
  #cat('NMVBSmix takes', end - begin, 'seconds\n\n')
  
  para = list(p=p, mu = mu , lambda=lambda, sigma = sigma)
  if(is.numeric(class)==T){
    true.clus=class
    km.clus=post.cate
    MCR=error.rate(km.clus, true.clus)
    ARI=adjustedRandIndex(post.cate, true.clus)
    jacc_ind=cluster_similarity(true.clus, post.cate)
    
    out=list(logli = logli.new, para = para, ni = di, post.cate = post.cate, 
             para.len = m,AIC = AIC, BIC = BIC, ICL = ICL,EDC = EDC , CLC = CLC,
             MCR = MCR)#, #Rand.index = RI[1], adjusted.Rand.Index =RI[2], jacc_ind = jacc_ind)
  }
  else{out = list(iter=iter ,logli = logli.new, para = para, ni = di, post.cate = post.cate, 
                  para.len = m,AIC = AIC, BIC = BIC, ICL = ICL,EDC = EDC , CLC = CLC)}
  mds=c(logli=logli.new, AIC=AIC, BIC=BIC,  ARI=ARI, CCR=1-MCR)
  return(list(mds=mds, para=para))
}
